##
#   Copyright 2021 Alibaba, Inc. and its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
##

# -*- coding: utf-8 -*-

from dashvector.common.types import *
from dashvector.core.models.partition_meta_status import PartitionMeta, PartitionStats
from dashvector.core.proto.dashvector_pb2 import CollectionInfo, StatsCollectionResponse


class CollectionMeta(object):
    def __init__(self, meta: Union[dict, CollectionInfo]):
        """
        meta: CollectionInfo
        """
        self._meta = None
        if isinstance(meta, CollectionInfo):
            self._parse_pb_collectioninfo(meta)
        elif isinstance(meta, dict):
            self._parse_dict_collectioninfo(meta)
        else:
            raise DashVectorException(
                code=DashVectorCode.InvalidArgument,
                reason="DashVectorSDK parse collection meta failed and meta type must be [dict, CollectionInfo]",
            )
        self._meta = meta

    def _parse_pb_collectioninfo(self, meta: CollectionInfo):
        self._name = meta.name
        self._dimension = meta.dimension
        self._dtype = meta.dtype
        self._metric = meta.metric
        self._status = meta.status

        # fields_schema
        self._fields = {}
        for field_name, field_type in meta.fields_schema.items():
            self._fields[field_name] = FieldType.str(field_type)

        # partitions
        self._partitions = {}
        for partition_name, partition_status in meta.partitions.items():
            self._partitions[partition_name] = PartitionMeta(partition_name, Status.str(partition_status))

    def _parse_dict_collectioninfo(self, meta: dict):
        self._name = meta.get("name")
        self._dimension = meta.get("dimension")
        self._dtype = VectorType.get(meta.get("dtype"))
        self._metric = MetricType.get(meta.get("metric"))
        self._status = Status.get(meta.get("status"))

        # fields_schema
        self._fields = meta.get("fields_schema")

        # partitions
        partitions = meta.get("partitions")
        if not bool(partitions):
            return
        self._partitions = {}
        for partition_name, partition_status in partitions.items():
            self._partitions[partition_name] = PartitionMeta(partition_name, partition_status)

    @property
    def name(self):
        return self._name

    @property
    def dimension(self):
        return self._dimension

    @property
    def dtype(self):
        return VectorType.str(self._dtype)

    @property
    def metric(self):
        return MetricType.str(self._metric)

    @property
    def fields_schema(self):
        return self._fields

    @property
    def status(self):
        return Status.str(self._status)

    @property
    def partitions(self):
        return self._partitions

    def __dict__(self):
        meta_dict = {"name": self.name, "dimension": self.dimension, "dtype": self.dtype, "metric": self.metric}
        if self._status is not None:
            meta_dict["status"] = self.status
        if self._fields is not None:
            meta_dict["fields_schema"] = self.fields_schema
        if self._partitions is not None:
            partitions_meta = {}
            for partition_name, partition_meta in self._partitions.items():
                partitions_meta[partition_name] = partition_meta.status
            meta_dict["partitions"] = partitions_meta
        return meta_dict

    def __str__(self):
        return json.dumps(self.__dict__())

    def __repr__(self):
        return self.__str__()


class CollectionStats(object):
    def __init__(self, *, stats: Union[dict, StatsCollectionResponse.CollectionStats]):
        """
        stats: Union[dict, StatsCollectionResponse.CollectionStats]
        """
        if stats is None:
            return
        self._stats = stats
        if isinstance(stats, StatsCollectionResponse.CollectionStats):
            self._parse_pb_collectionstats()
        elif isinstance(stats, dict):
            self._parse_dict_collectionstats()
        else:
            raise DashVectorException(
                code=DashVectorCode.InvalidArgument,
                reason="DashVectorSDK parse collection stats failed and stats type must be [dict, CollectionStats]",
            )

    def _parse_pb_collectionstats(self):
        self._total_doc_count = self._stats.total_doc_count
        self._index_completeness = round(self._stats.index_completeness, 1)
        self._partitions = {}
        for partition_name, partition_stats in self._stats.partitions.items():
            self._partitions[partition_name] = PartitionStats(
                name=partition_name, total_doc_count=partition_stats.total_doc_count
            )

    def _parse_dict_collectionstats(self):
        self._total_doc_count = self._stats.get("total_doc_count")
        self._index_completeness = round(self._stats.get("index_completeness"), 1)
        partitions = self._stats.get("partitions")
        if not bool(partitions):
            return
        self._partitions = {}
        for partition_name, partition_stats in partitions.items():
            self._partitions[partition_name] = PartitionStats(
                name=partition_name, total_doc_count=partition_stats.get("total_doc_count")
            )

    @property
    def total_doc_count(self):
        return self._total_doc_count

    @property
    def index_completeness(self):
        return self._index_completeness

    @property
    def partitions(self):
        return self._partitions

    def __dict__(self):
        meta_dict = {"total_doc_count": self.total_doc_count, "index_completeness": self.index_completeness}
        partitions = {}
        if len(self._partitions) > 0:
            for partition_name, partition_stats in self._partitions.items():
                partitions[partition_name] = partition_stats.__dict__()
            meta_dict["partitions"] = partitions
        return meta_dict

    def __str__(self):
        return json.dumps(self.__dict__())

    def __repr__(self):
        return self.__str__()
